

import calendar
import copy
from datetime import datetime, timedelta
import hashlib
import logging
import os
import socket
import subprocess as sp
import sys

import fantail
import humanize
from leip import set_local_config

lg = logging.getLogger(__name__)


BASEUID = None
def get_base_uid():
    global BASEUID
    if BASEUID is None:
        sha1 = hashlib.sha1()
        sha1.update(str(datetime.now()))
        sha1.update(socket.getfqdn())
        BASEUID = sha1.hexdigest()[:8]
    return BASEUID
    
def get_uid(app, runno=None):
    if app.args.uid:
        base = app.args.uid
    else:
        base = get_base_uid()
        
    if runno:
        return '{}.{}'.format(base, runno)
    else:
        return base

        
# Nicely format run log stats
FSIZEKEYS = ["ps_meminfo_max_rss", "ps_meminfo_max_vms",
             "ps_sys_swap_free", "ps_sys_swap_sin",
             "ps_sys_swap_sout", "ps_sys_swap_total",
             "ps_sys_swap_used", "ps_sys_vmem_active",
             "ps_sys_vmem_available", "ps_sys_vmem_buffers",
             "ps_sys_vmem_cached", "ps_sys_vmem_free",
             "ps_sys_vmem_inactive", "ps_sys_vmem_total",
             "ps_sys_vmem_used"]


def nicetime(t):
    """
    fomat time, assuming t is in utc
    """

    timestamp = calendar.timegm(t.timetuple())
    loct = datetime.fromtimestamp(timestamp)
    assert t.resolution >= timedelta(microseconds=1)
    loct.replace(microsecond=t.microsecond)
    return '{} ({})'.format(humanize.naturaltime(loct), t)

    
def make_pretty_kv(k, v):
    if k in ['cl', 'template_cl']:
        v = " ".join(v)
            
    if v is None:
        return ""
    elif isinstance(v, dict):
        return v
    elif str(v).strip() == "":
        return ""
    if k in FSIZEKEYS:
        return "{} ({})".format(humanize.naturalsize(v), v)
    elif k in ['start', 'stop']:
        return nicetime(v)
    elif k == 'runtime':
        return '{}s ({})'.format(humanize.intword(v), v)
    elif k.endswith('_percent'):
        return '{}%'.format(v)
    else:
        return str(v)

def make_pretty_kv_html(k, v):
    
    if k in ['cl', 'template_cl']:
        return (
            '<td colspan="2">' +
            '<span style="font-family:Lucida Console,Bitstream Vera Sans Mono,Courier New,monospace;">' +
            " ".join(v) +
            "</span></td>" )
            
    if v is None:
        return '<td colspan="2"></td>'
    elif str(v).strip() == "":
        return '<td colspan="2"></td>'
        
    if k == 'files':
        rv = ['<td colspan="2"><table>']
        for fn in v:
            fd = " ".join(['/'.join(x) for x in v[fn]])
            rv.append('<tr><td><b>{}</b></td>'.format(fn))
            
            rv.append('<td>: <i>{}</i></td>'.format(fd))
            rv.append('</tr>')
            
        rv.append('</table></td>')
        return "".join(rv)
        
    if k in FSIZEKEYS:
        return "<td>{}</td><td>{}</td>".format(humanize.naturalsize(v), v)
    elif k in ['start', 'stop']:
        return '<td colspan="2">{}</td>'.format(v)
    elif k == 'runtime':
        return "<td>{}</td><td>{}</td>".format(humanize.intword(v), v)
    elif k.endswith('_percent'):
        return '<td colspan="2">{}%</td>'.format(v)
    else:
        return '<td colspan="2">{}</td>'.format(v)

def get_tool_conf(app, name, version='default'):

    data = copy.copy(app.conf['group.default'])
    tool_data = copy.copy(app.conf.get('app.{}'.format(name), fantail.Fantail()))
    group = tool_data.get('group')

    if not group is None:
        group_data = app.conf['group.{}'.format(group)]
        if group_data:
            data.update(group_data)

    data.update(tool_data)

    if version is 'default':
        version = tool_data.get('default_version', None)

    if (not version is None) and (not version in tool_data['versions']):
        candidates = []
        for v in tool_data['versions']:
            fullv =  tool_data['versions'][v]['version']
            if v in fullv:
                candidates.append(v)

    if not version is None:
        version_data = tool_data['versions.{}'.format(version)]
        data.update(version_data)
        data['version_key'] = version

    return data


def is_kea(fname):

    with open(fname) as F:
        start = F.read(1000)

    fline = start.strip().split("\n")[0]
    if not fline.startswith('#!'):
        lg.debug(" - not a shell script - not kea")
        return False

    if not 'python' in fline:
        lg.debug(" - not a python script - not kea")
        return False

    if 'load_entry_point' in start and \
            'Kea==' in start:
        lg.debug(" - looks like a link to the kea entry point script - kea")
        return True

    if 'import Kea' in start or \
            'from Kea import' in start:
        lg.debug(" - looks like custom Kea script - kea")
        return True

    lg.debug(" - does not look like a kea script")
    return False


def find_executable(name):

    # check if this is a single executable:
    if os.path.isfile(name) and os.access(name, os.X_OK):
        executable = name
        name = os.path.basename(executable)
        yield os.path.abspath(executable)

    else:

        # no? try to use the 'which' tool

        # no '/' allowed anymore
        if '/' in name:
            raise IOError(name)

        P = sp.Popen(['which', '-a', name], stdout=sp.PIPE)

        out, err = P.communicate()

        for line in out.strip().split("\n"):
            lg.debug("check %s", line)
            if not is_kea(line):
                lg.debug("%s is not a kea file", line)

                yield os.path.abspath(line)


def create_kea_link(app, name):
    """
    """
    base = app.conf['bin_path']
    linkpath = os.path.expanduser(os.path.join(base, name))
    lg.debug("checking: %s", linkpath)
    if os.path.lexists(linkpath):
        lg.debug("path exists: %s", linkpath)
        os.unlink(linkpath)


    keapath = sys.argv[0]
    lg.info("creating link from %s", linkpath)
    lg.info(" to: %s", keapath)
    os.symlink(keapath, linkpath)

def register_executable(app, name, executable, version, is_default=None):
    """
    Register an executable
    """

    allversions = list('abcdefghijklmnopqrstuvwxyz123456789')

    is_first_version = True

    version_key = 'a'

    if app.conf.has_key('app.{}.versions'.format(name)):
        is_first_version = False
        for k in app.conf['app.{}.versions'.format(name)]:
            vinf = app.conf['app.{}.versions.{}'\
                            .format(name, k)]
            if vinf['executable'] == executable:
                lg.warning("Executable is already registered - overwriting")
                version_key = k
                break

            #registered - we do not want to use this key
            allversions.remove(k)

        version_key = allversions[0]

    if is_default == False:
        if is_first_version:
            lg.debug("First version of %s - setting to default", name)
            is_default = True
        else:
            lg.debug("Other version of %s present - not setting default", name)
            is_default = False

    lg.warning("register %s - %s - %s - %s", name, executable,
             version_key, version)

    if is_default:
        lg.warning("Set version %s as default", version_key)
        set_local_config(app, 'app.{}.default_version'.format(name),
                         version_key)

    basekey = 'app.{}.versions.{}'.format(name, version_key)
    lg.debug("register to: %s", basekey)

    set_local_config(app, '{}.executable'.format(basekey), executable)
    set_local_config(app, '{}.version'.format(basekey), version)

    create_kea_link(app, name)




import leip
import sys
import subprocess as sp

from mad2.recrender import recrender
#from mad2.plugin.checksum import apply_checksum_madfile

# @leip.hook('finish')
# def annotate_provenance(app):

#     if not app.conf['output_files']:
#         #no output files - nothing to annotate - move on
#         return

#     #sha1:
#     for a in app.conf['output_files']:
#         apply_checksum_madfile({}, a, 'sha1')

#     # for tool in app.conf['annotate']:
#     #     tinfo = app.conf['annotate'][tool]
#     #     key = tinfo['key']
#     #     key = recrender(key, app.conf)
#     #     val = None
#     #     if 'value' in tinfo:
#     #         val = tinfo['value']
#     #         val = recrender(val, app.conf)
#     #     elif 'command' in tinfo:
#     #         comm = tinfo['command']
#     #         comm = recrender(comm, app.conf)
#     #         val = sp.Popen(comm, shell=True,
#     #                        stdout=sp.PIPE).communicate().strip()
#     #     if val:
#     #         kvpairs.append((key, val))

#     # print kvpairs
#     # # if 'provenance.version_command' in app.conf:
#     # #     vcm = app.conf['provenance.version_command']
#     # #     vcm = recrender(vcm, app.conf)
#     # #     version, _ = sp.Popen(vcm, shell=True, stdout=sp.PIPE).communicate()
#     # #     version = version.strip()

#     # cl.append('-g')
#     # cl.append("{}".format(app.conf['appname']))
#     # cl.append("{}".format(app.conf['executable']))
#     # cl.append("{}".format('1'))

#     # for i in app.conf['input_files']:

#     #     cl.append('--derived_from')
#     #     cl.append(i)

#     # for o in app.conf['output_files']:
#     #     cl.append(o)
#     # #print " ".join(cl)
#     # if app.conf.get('command_echo'):
#     #     print " ".join(cl)

#     #sp.Popen(cl).communicate()




from collections import deque
import copy
from datetime import datetime
import fcntl
import hashlib
import logging
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing.dummy import Lock
import os
import re
import shlex
import subprocess as sp
import sys
import tempfile
import time


import psutil

import leip

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)
MPJOBNO = 0

@leip.hook('pre_argparse')
def main_arg_define(app):
    if app.executor == 'simple':
        simple_group = app.parser.add_argument_group('Simple Executor')
        simple_group.add_argument('-T', '--no_track_stat', help='do not track process status',
                                action='store_true', default=None)
        simple_group.add_argument('-j', '--threads', help='no threads to use',
                                type=int)
        simple_group.add_argument('-E', '--echo', help='echo command line to screen',
                                action='store_true', default=None)
        simple_group.add_argument('-w', '--walltime',
                                help=('max time that this process can take, ' +
                                      'after this time, the process gets killed. ' +
                                      'Specified in seconds, or with ' +
                                      'postfix m for minutes, h for hours, ' +
                                      'd for days'))
                


#thanks: https://gist.github.com/sebclaeys/1232088
def non_block_read(stream, chunk_size=10000):
    fd = stream.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
    try:
        return stream.read()
    except:
        return ""

outputlock = Lock()

BINARY_STREAMS = set()

def streamer(src, tar, dq, hsh=None):
    """
    :param src: input stream
    :param tar: target stream
    :param dq: deque object keeping a tail of chunks for this stream
    :param hsh: hash object to calculate a checksum 
    """
    d = non_block_read(src)
    if d is None:
        return 0
    if hsh:
        hsh.update(d)
        
    global BINARY_STREAMS
    stream_id = '{}_{}'.format(src.__repr__(), tar.__repr__())

    if not stream_id in BINARY_STREAMS:
        try:
            dq.append(d.decode('utf-8'))
        except UnicodeDecodeError:
            BINARY_STREAMS.add(stream_id)
        
    d_len = len(d)
    with outputlock:
        tar.write(d) #d.encode('utf-8'))
    return d_len

    
def get_deferred_cl(info):
    dcl = ['kea']
    if info['stdout_file']:
        dcl.extend(['-o', info['stdout_file']])
    if info['stderr_file']:
        dcl.extend(['-e', info['stderr_file']])
    dcl.extend(info['cl'])
    return dcl

    
def store_process_info(info):
    psu = info.get('psutil_process')
    if not psu: return

    try:
        info['pid'] = psu.pid
        info['ps_nice'] = psu.nice()
        info['ps_num_fds'] = psu.num_fds()
        info['ps_threads'] = psu.num_threads()

        cputime = psu.cpu_times()

        info['ps_cputime_user'] = cputime.user 
        info['ps_cputime_system'] = cputime.system
        info['ps_cpu_percent_max'] = max(
            info.get('ps_cpu_percent_max', 0),
            psu.cpu_percent())

        meminfo = psu.memory_info()

        for f in meminfo._fields:
            info['ps_meminfo_max_{}'.format(f)] = \
                    max(getattr(meminfo, f),
                        info.get('ps_meminfo_max_{}'.format(f), 0))

        try:
            ioc = psu.io_counters()
            info['ps_io_read_count'] = ioc.read_count
        except AttributeError:
            #may not have iocounters (osx)
            pass

    except (psutil.NoSuchProcess, psutil.AccessDenied):
        #process went away??
        return

def simple_runner(info, executor, defer_run=False):
    """
    Defer run executes the run with the current executor, but with
    the Kea executable so that all kea related functionality is
    executed in the second stage.
    """

    #get a thread run number
    global MPJOBNO
    thisjob = MPJOBNO
    MPJOBNO += 1

    #track system status (memory, etc)
    sysstatus = not executor.app.defargs['no_track_stat']
    
    info['job_thread_no'] = MPJOBNO
    lgx = logging.getLogger("job{}".format(thisjob))
    #lgx.setLevel(logging.DEBUG)
    
    stdout_handle = sys.stdout  # Unless redefined - do not capture stdout
    stderr_handle = sys.stderr  # Unless redefined - do not capture stderr

    walltime = executor.walltime
    
    if defer_run:
        cl = get_deferred_cl(info)
    else:
        cl = info['cl']
        if info['stdout_file']:
            lg.debug('capturing stdout in %s', info['stdout_file'])
            stdout_handle = open(info['stdout_file'], 'w')
        if info['stderr_file']:
            lg.debug('capturing stderr in %s', info['stderr_file'])
            stderr_handle = open(info['stderr_file'], 'w')

    info['start'] = datetime.utcnow()
    lgx.debug("thread start %s", info['start'])

    #system psutil stuff
    if sysstatus:
        info['ps_sys_cpucount'] = psutil.cpu_count()
        psu_vm = psutil.virtual_memory()
        for field in psu_vm._fields:
            info['ps_sys_vmem_{}'.format(field)] = getattr(psu_vm, field)
        psu_sw = psutil.swap_memory()
        for field in psu_sw._fields:
            info['ps_sys_swap_{}'.format(field)] = getattr(psu_sw, field)


    if defer_run:
        P = psutil.Popen(cl, shell=True)
        info['pid'] = P.pid
        info['submitted'] = datetime.utcnow()
        return info

    #execute!
    lgx.debug("Starting: %s", " ".join(cl))
    strace_file = tempfile.NamedTemporaryFile(delete=False)
    strace_file.close()
    mcl = "strace -e trace=file -tt -r -f -o {} ".format(strace_file.name)
    lg.debug("strace to: %s", strace_file.name)
    mcl += " ".join(cl) 
    lg.debug("executing: %s", mcl)
    P = psutil.Popen(mcl, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)

    info['strace_pid'] = P.pid
    lgx.debug("Job has started with (strace) pid: %d", info['strace_pid'])

    def _get_psu(pid):
        ppsu = psutil.Process(pid)
        child = ppsu.children()
        if len(child) == 0:
            return False
        assert len(child) == 1
        return child[0]

    if sysstatus:
        try:
            psu = _get_psu(P.pid)
            if isinstance(psu, psutil.Process):
                info['psutil_process'] = psu
                store_process_info(info)
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            #job may have already finished - ignore
            lg.info('has the job finished already??')
            pass

    stdout_dq = deque(maxlen=100)
    stderr_dq = deque(maxlen=100)

    stdout_len = 0
    stderr_len = 0

    stdout_sha = hashlib.sha1()
    stderr_sha = hashlib.sha1()

    force_quit = False

    #loop & poll until the process finishes..
    while True:

        if walltime:
            runtime = (datetime.utcnow() - info['start']).total_seconds()
            if runtime > walltime:
                if not force_quit:
                    lg.warning("Killing process (walltime)")
                force_quit = True
                P.kill()

        pc = P.poll()

        if isinstance(pc, int):
            #integer <- returncode => so exit.
            lgx.debug("got a return code (%d) - exit", pc)
            break

        #read_proc_status(td)
        time.sleep(0.2)
        if sysstatus:
            if 'psutil_process' in info:
                store_process_info(info)
            else:
                psu = _get_psu(P.pid)
                if isinstance(psu, psutil.Process):
                    info['psutil_process'] = psu
                    store_process_info(info)
            


        try:
            stdout_len += streamer(P.stdout, stdout_handle, stdout_dq, stdout_sha)
            stderr_len += streamer(P.stderr, stderr_handle, stderr_dq, stderr_sha)
        except IOError as e:
            #it appears as if one of the pipes has failed.
            if e.errno == 32:
                errors.append("Broken Pipe")
                #broken pipe - no problem.
            else:
                message('err', str(dir(e)))
                errors.append("IOError: " + str(e))
            break

    #clean the pipes
    try:
        stdout_len += streamer(P.stdout, stdout_handle, stdout_dq, stdout_sha)
        stderr_len += streamer(P.stderr, stderr_handle, stderr_dq, stderr_sha)
    except IOError as e:
        #it appears as if one of the pipes has failed.
        if e.errno == 32:
            errors.append("Broken Pipe")
            #broken pipe - no problem.
        else:
            message('err', str(dir(e)))
            errors.append("IOError: " + str(e))


    if force_quit:
        info['status'] = 'killed'
    elif P.returncode == 0:
        info['status'] = 'success'
    else:
        info['status'] = 'fail'

    if info['stdout_file']:
        lg.debug('closing stdout handle on %s', info['stdout_file'])
        stdout_handle.close()
    if info['stderr_file']:
        lg.debug('closing stderr handle on %s', info['stderr_file'])
        stderr_handle.close()

    lgx.debug("job has status: %s", info['status'])

    if stdout_len > 0:
        info['stdout_sha1'] = stdout_sha.hexdigest()
    if stderr_len > 0:
        info['stderr_sha1'] = stderr_sha.hexdigest()

    info['stop'] = datetime.utcnow()
    info['runtime'] = (info['stop'] - info['start']).total_seconds()

    if 'psutil_process' in info:
        del info['psutil_process']

    info['returncode'] = P.returncode
    info['stdout_len'] = stdout_len
    info['stderr_len'] = stderr_len
    lgx.debug("end thread")


    #parse strace output & store data
    if not 'files' in info:
        info['files'] = {}

    def _fileupdate(d, k, v):
        if not k in d:
            d[k] = []
        if not v in d[k]:
            d[k].append(v)

    to_ignore = [
        re.compile(r'/site-packages/'),
        re.compile(r'/project/Mad2/'),
        re.compile(r'/project/Leip/'),
        re.compile(r'/lib/python'),
        re.compile(r'/project/Fantail/'),
        re.compile(r'^/dev/'),
        re.compile(r'^/proc/'),
        re.compile(r'^/lib/'),
        re.compile(r'^/usr/lib/'),
        re.compile(r'^/usr/local/lib/'),
    ]
            
    with open(strace_file.name) as F:
        for line in F:
            line = line.strip()
            ls = line.split(None, 2)
            if '(No such file or directory)' in line:
                continue
            typ, rest = ls[2].split('(',1)
            rest = shlex.split(rest.rsplit(')',1)[0])
            filename = rest[0].rstrip(',')
            if os.path.isdir(filename):
                continue #no directories
            ignore = False
            for toi in to_ignore:
                if toi.search(filename):
                    ignore=True
                    break
            if ignore:
                continue
            if 'stat' in typ: continue
            if 'SIGCHLD' in typ and filename == 'Child': 
                continue
            if 'exec' in typ:
                _fileupdate(info['files'], filename,
                            [typ])
            else:
                _fileupdate(info['files'], filename, 
                            [typ] + rest[1:])
    os.unlink(strace_file.name)


    return info
        
class BasicExecutor(object):

    def __init__(self, app):
        lg.debug("Starting executor")
        self.app = app
        
        try:
            self.threads =  self.app.args.threads
        except AttributeError:
            self.threads = 1
            
        if self.threads < 2:
            self.simple = True
        else:
            self.simple = False
            self.pool = ThreadPool(self.threads)
            lg.debug("using a threadpool with %d threads", self.threads)


        self.walltime = None
        if hasattr(self.app.args, 'walltime'):
            w = self.app.args.walltime
            if not w is None:
                if len(w) > 1 and w[-1] == 'm':
                    self.walltime = float(w[:-1]) * 60
                elif len(w) > 1 and w[-1] == 'h':
                    self.walltime = float(w[:-1]) * 60 * 60
                elif len(w) > 1 and w[-1] == 'd':
                    self.walltime = float(w[:-1]) * 60 * 60 * 24
                else:
                    self.walltime = float(w)
                
    def fire(self, info):
        lg.debug("start execution")
        
        self.app.run_hook('pre_fire', info)
        
        if self.app.args.echo:
            print(" ".join(info['cl']))

            
        if self.simple:
            simple_runner(info, self)
            self.app.run_hook('post_fire', info)
        else:
            def _callback(info):
                lg.warningdebug("job finished (%s) - callback!", info['job_thread_no'])
                self.app.run_hook('post_fire', info)

            self.pool.apply_async(simple_runner, [info, self], {'defer_run': False},
                                  callback=_callback)

    def finish(self):
        if not self.simple:
            lg.info('waiting for the threads to finish')
            self.pool.close()
            self.pool.join()
            lg.debug('finished waiting for threads to finish')


class DummyExecutor(BasicExecutor):

    def fire(self, info):
        
        self.app.run_hook('pre_fire', info)
        lg.debug("start dummy execution")
        cl = copy.copy(info['cl'])

        if info['stdout_file']:
            cl.extend(['>', info['stdout_file']])
        if info['stderr_file']:
            cl.extend(['2>', info['stderr_file']])

        lg.debug("  cl: %s", cl)
        print " ".join(cl)
        info['mode'] = 'synchronous'
        info['returncode'] = 0
        info['status'] = 'success'
        self.app.run_hook('post_fire', info)
        

conf = leip.get_config('kea')
conf['executors.simple'] = BasicExecutor
conf['executors.dummy'] = DummyExecutor




#!/usr/bin/env python

import argparse
import collections
import glob
import imp
import logging
import os
from pprint import pprint
import sys

import yaml


#import grako.exception
parser = argparse.ArgumentParser()
parser.add_argument('-x', action='store_true')
parser.add_argument('names', nargs='*')

args = parser.parse_args()


logging.basicConfig(level=logging.INFO)
lg = logging.getLogger(__name__)

def runtest(x):

    base = x.replace('.test', '')
    
    lg.info("loading module %s", base)
    modinf = imp.find_module(base)
    mod = imp.load_module(base, *modinf)
    parser = getattr(mod, '{}Parser'.format(base))()

    def lineparser(F):
        line = F.readline()
        nextline = F.readline()

        while True:
            
            if nextline[0] == '>':
                
                tdata = yaml.load(nextline[1:].strip())
                def listifykeys(d):
                    for k, v in d.items():
                        if isinstance(v, dict):
                            listifykeys(v)
                        elif not isinstance(v, list):
                            d[k] = [v]
                listifykeys(tdata)
                yield line.strip(), tdata
                nextline = F.readline()
            else:
                yield line.strip(), None
            
            if not nextline:
                break
                
            line = nextline
            nextline = F.readline()

        if nextline is None and line:
            yield line, None
        elif nextline:
            yield nextline, None
                

    #thanks: http://tinyurl.com/nttpj9
    def convert_to_str(data):
        if isinstance(data, basestring):
            return str(data)
        elif isinstance(data, collections.Mapping):
            return dict(map(convert_to_str, data.iteritems()))
        elif isinstance(data, collections.Iterable):
            return type(data)(map(convert_to_str, data))
        else:
            return data

    def astparse(ast, data={}):
        if isinstance(ast, list):
            [astparse(x, data) for x in ast]
        elif isinstance(ast, dict):
            data.update(convert_to_str(ast))
        elif isinstance(ast, (str, unicode)):
            pass
        else:
            print('x' * 80, ast)
        return data


    with open(x) as F:

        for line, test in lineparser(F):
            
            if test is None:
                tres = '.'

            mustfail = '+'
            if line[0] == '-':
                #must fail!
                mustfail = '-'
                line = line[1:].strip()

            if line[0] == '#': continue

            try:
                res = parser.parse(line, rule_name="start")
                success = '+'
            except Exception as e: #grako.exception.FailedParse as e:
                if args.x and mustfail != '-':
                    raise
                success = '-'
                
            if success == '+':
                data = astparse(res, data = {})
                if not test is None:
                    if test == data:
                        tres = '+'
                    else:
                        tres = '-'


            testresult = '+:' if success == mustfail else '-:'
            if tres == '-':
                testresult = '-'

            print "{}{}{}{} : {}".format(testresult, mustfail, success, tres, line)

            if not '+' in testresult and success == '+':
                pprint(data)
            if tres == '-':
                pprint(test)
            



if not args.names:
    for x in glob.glob('*.test'):
        runtest(x)
else:
    for x in args.names:
        runtest(x)
    




import os
import re

import leip

import toMaKe.param

@leip.hook('prepare')
def harvest_params(app):
    app.param = toMaKe.param.get_recursive_param(os.getcwd(), 'kea.config')
    
    
@leip.hook('pre_fire')
def render_param(app, info):
    cl = info['cl']
    findpar = re.compile(r'{([a-zA-Z_][a-zA-Z0-9_]*)}')

    for i, argument in enumerate(cl):
        while True:
            done = False
            for fp in findpar.finditer(argument):
                nm = fp.groups()[0]
                repl = app.param[nm]
                
                if not repl:
                    #do not have a value for this variable - continue
                    continue
                    
                argument = findpar.sub(app.param[nm], argument)
                #have  a value - restart, because 'argument' has changed
                break
            else:
                #nothing new found, nothing to replace then, done
                done=True
                break

            if done:
                break

        #replace argument in commandline
        cl[i] = argument

    #replace command line
    info['cl'] = cl


        
#    print(cl)



"""
Map arguments

input:

    {name~*}

    the * is treated as a glob and can be part of a filename
    name is used to expand the capture part later.

    {name=[1-5]}

    name is replaced by the range of

    {name=a,b,c,d}

"""

import copy
from datetime import datetime
import hashlib
import glob
import logging
import itertools
import re
import shlex
import socket
import sys
from collections import OrderedDict

import leip
from kea.utils import get_uid

lg = logging.getLogger(__name__)



RE_FIND_MAPINPUT = re.compile(r'{([a-zA-Z_][a-zA-Z0-9_]*)([\~\=])([^}]+)}')


def map_range_expand(map_info, cl, pipes):
    """
    Convert a map range to a list of items:

    first - range conversion:
      - a,b,c -> ['a', 'b', 'c']
      - 5:10 -> [5,6,7,8,9]
      - 5:10:2 -> [5,7,9]
    then, merge with the command line item in which this was embedded (if any)

    so: test.{a=5:10:2}.in becomes: ['test.5.in', 'test.7.in', 'test.9.in']

    """

    mappat_range_3 = re.match(r'([0-9]+):([0-9]+):([0-9]+)',
                              map_info['pattern'])
    mappat_range_2 = re.match(r'([0-9]+):([0-9]+)',
                              map_info['pattern'])

    thisarg = map_info['arg']
    iterstring = map_info['iterstring']
    argi = cl.index(thisarg)

    if mappat_range_3:
        start, stop, step = mappat_range_3.groups()
        map_items = map(str, range(int(start), int(stop), int(step)))
    elif mappat_range_2:
        start, stop = mappat_range_2.groups()
        map_items = map(str, range(int(start), int(stop)))
    elif ',' in map_info['pattern']:
        map_items = [x.strip() for x in map_info['pattern'].split(',')]
    else:
        lg.critical("Can not parse range: %s", map_info['pattern'])
        exit(-1)

        
    substart = thisarg.index(iterstring)
    subtail = substart + len(iterstring) 
    for g in map_items:
        newcl = copy.copy(cl)
        argrep = thisarg[:substart] + str(g) +  thisarg[subtail:]
        newcl[argi] = argrep
        for j, rarg in enumerate(newcl):
            newcl[j] = map_info['rep_from'].sub(g, rarg)


        new_pipes = []
        for p in pipes:
            if p is None:
                new_pipes.append(None)
            else:
                new_pipes.append(map_info['rep_from'].sub(g, p))

        yield newcl, new_pipes


def map_glob_expand(map_info, cl, pipes):
    thisarg = map_info['arg']
    argi = cl.index(thisarg)
    iterpat = map_info['pattern']
    iterstring = map_info['iterstring']

    globpattern = RE_FIND_MAPINPUT.sub(iterpat, thisarg)
    globhits = glob.glob(globpattern)

    if len(globhits) == 0:
        lg.critical("No files matching pattern: '%s' found", globpattern)
        exit(-1)

    substart = thisarg.index(iterstring)
    subtail = len(thisarg) - (substart + len(iterstring))
    
    for g in globhits:
        newcl = copy.copy(cl)
        newcl[argi] = g
        subrep = g[substart:]
        if subtail > 0:
            subrep = subrep[:-subtail]
        for j, rarg in enumerate(newcl):
            newcl[j] = map_info['rep_from'].sub(subrep, rarg)

        new_pipes = []
        for p in pipes:
            if p is None:
                new_pipes.append(None)
            else:
                new_pipes.append(map_info['rep_from'].sub(g, p))

        yield newcl, new_pipes


def map_iter(map_info):
    for i in map_info['items']:
        map_clean = copy.copy(map_info)
        map_clean['item'] = i
        yield map_clean


def apply_map_info_to_cl(newcl, map_info):
    item = map_info['item']
    for i, arg in enumerate(newcl):
        map_to_re = map_info['re_replace']
        map_fr_re = map_info['re_from']
        if map_fr_re.search(arg):
            newcl[i] = map_fr_re.sub(item, arg)
        elif map_to_re.search(arg):
            newcl[i] = map_to_re.sub(item, arg)

    return newcl

    
    
def basic_command_line_generator(app):
    """
    Most basic command line generator possible
    """
    
    info = OrderedDict()
    
    pipes = [app.args.stdout, app.args.stderr]

    cl = app.cl

    cljoin = " ".join(cl)

    #check if there are iterable arguments in here
    mapcount = 0
    mapins = RE_FIND_MAPINPUT.search(cljoin)
    nojobstorun = app.defargs['jobstorun']

    
    # no map definitions found - then simply return the cl & execute
    if mapins is None:
        info['cl'] = cl
        info['run_no'] = 0
        info['run_uid'] = get_uid(app)
        info['stdout_file'] = pipes[0]
        info['stderr_file'] = pipes[1]
        yield info
        return

    info['template_cl'] = cl
    #define iterators for each of the definitions

    #lg.setLevel(logging.DEBUG)

    lg.debug("iterables found in: %s", cljoin)
    
    def expander(cl, pipes):
        
        for i, arg in enumerate(cl):
            mima = RE_FIND_MAPINPUT.search(arg)
            if mima:
                break
        else:
            yield cl, pipes
            return

        map_info = {}
        map_info['re_search'] = mima
        map_info['name'], map_info['operator'], map_info['pattern'] = \
                mima.groups()

        map_info['iterstring'] = mima.group(0)
        map_info['arg'] = arg
        

        lg.debug("iterable name: {name} operator: {operator} pattern: {pattern}".format(**map_info))
        
        map_info['re_from'] = re.compile(r'({' + map_info['name']
                                         + r'[\~\=][^}]*})')
        map_info['re_replace'] = re.compile(r'({' + map_info['name'] + r'})')
        map_info['rep_from'] = re.compile(r'({' + map_info['name'] + '})')
        
        map_info['start'] = mima.start()
        map_info['tail'] = len(arg) - mima.end()

        if map_info['operator'] == '~':
            expand_function = map_glob_expand
        elif map_info['operator'] == '=':
            expand_function = map_range_expand
        
        for ncl, pipes in  expand_function(map_info, cl, pipes):
            for nncl, pipes in expander(ncl, pipes):
                    yield nncl, pipes

    no = 0
    for newcl, pipes in  expander(cl, pipes):
        no += 1
        if nojobstorun and no > nojobstorun:
            break
        newinfo = copy.copy(info)
        newinfo['run_uid'] = get_uid(app, no)
        newinfo['cl'] = newcl
        newinfo['run_no'] = no
        newinfo['stdout_file'] = pipes[0]
        newinfo['stderr_file'] = pipes[1]
        yield newinfo



"""
Send mail plugin


"""

import leip
import logging
import subprocess as sp

import jinja2

import kea.utils

lg = logging.getLogger(__name__)


@leip.hook('pre_argparse')
def define_args(app):
    mail_group = app.parser.add_argument_group('Mail plugin')

    mail_group.add_argument('--mail_send', help='Send an email report',
                            action='store_true', default=None)
    
    mail_group.add_argument('-m', '--mail_on_success', 
                            help='Also send an email when a job finishes succesfully',
                            action='store_true', default=None)

    mail_group.add_argument('--mail_recipient', 
                            help='email address to send report to (comma ' +
                            'separate multiple addresses)',
                            action='store_true', default=None)

        
HTML_MESSAGE = """MIME-Version: 1.0
Content-Type: text/html;
Subject: 
{%- if success %} Kea/Ok: {{cl50}} 
{%- else %} Kea/Fail: {{ cl50 }} 
{%- endif %}


{% if success -%}
<h2 style='color:green'>Kea/Job OK</h2>
{%- else -%}
<h2 style='color:red'>Kea/Job FAIL</h2>
{%- endif %}

{% if all_jinf|length > 1 -%}
<h3>Template Command Line</h3>
<pre>
  {{ clj }}
</pre>

<h3>Actual command lines</h3>
{%- else -%}
<h3>Command line</h3>{% endif %}

<pre>{% for jinf in all_jinf -%}
{{ jinf.run_no }}.  {{ jinf['cl']|join(" ") }}

{% endfor %}
</pre>
{%- if all_jinf|length > 1 %}
<h3>Run stats</h3>
 
<table>
<tr>
  <th>No.</th>
{% if not success %}
  <th>rc</th>
{% endif %}
  <th>Run time</th>
  <th>Cpu time</th>
  <th>Peak mem</th>
  <th>Stdout</th>
  <th>Stderr</th>
</tr>
{% for jinf in all_jinf %}
<tr>
  <td><b>{{ jinf.run_no }}</b></td>
  {%- if not success %}{% if jinf.returncode == 0 %}
  <td><span style='color:green'>0</span></td>
  {% else %}
  <td><span style='color:red;weight:bold;'>{{jinf.returncode}}</span></td>
  {%- endif %}{% endif %}
  <td>{{ jinf.runtime }}</td>
  <td>{{ "%.3f"|format(jinf.ps_cputime_user) }}</td>
  <td>{{ "%d"|format(jinf.ps_meminfo_max_rss) }}</td>
  <td>{{ jinf.stdout_len }}</td>
  <td>{{ jinf.stderr_len }}</td>
</tr>
{%- endfor %}
</table>
{% endif %}

{% for jinf in all_jinf %}
<h3>Full run report {% if all_jinf|length > 1 %} ({{ jinf.run_no }}) {% endif %}</h3>
<table>
{% for k in jinf %}
<tr {% if loop.cycle(False, True) -%}style="background-color: #EEEEEE;"
    {%- endif %}><th style="text-align: left; vertical-align: text-top;">{{k}}</th>
    {{ jinf[k]|pretty(k) }}
</tr>
{%- endfor %}
</table>

{% endfor %}

"""

@leip.hook('finish')
def mail(app):

    #lg.setLevel(logging.DEBUG)
    if not hasattr(app, 'all_jinf'):
        return

    maildef = app.defargs.get('send_mail')
    mailsuc = app.defargs.get('mail_on_success')
    mailrecip = app.defargs.get('mail_recipient')


    if not maildef:
        lg.debug("no mail to be send")
        #sending mail is not required
        return

    #did all jobs finish successfully?
    success = True
    
    for jinf in app.all_jinf:
        if jinf['returncode'] != 0:
            success = False
            break

    if not mailsuc and success:
        lg.debug("job finished successfully - not sending mail")
        return
        
    data = {}
    data['success'] = success
    data['all_jinf'] = app.all_jinf
    data['app'] = app
    data['clj50'] = " ".join(app.cl)[:50]
    data['clj'] = " ".join(app.cl)

    
    if not mailrecip:
        lg.debug("No mail recipient(s) defined - cannot send mail")
        return

    def keapretty(value, key):
        return kea.utils.make_pretty_kv_html(key, value)

    env = jinja2.Environment()
    env.filters['pretty'] = keapretty
    template = env.from_string(HTML_MESSAGE)
    
    message = template.render(data)

    if ',' in mailrecip:
        mailrecip = " ".join(mailrecip.split(','))

    lg.debug("Sending report mail to: %s", mailrecip)
                             
    p = sp.Popen("sendmail %s" % mailrecip, stdin=sp.PIPE, shell=True)
    p.communicate(message.encode('utf-8'))


    exit(p.returncode)




import argparse
import logging
import os
import shlex
import sys

import leip

from kea import Kea

lg = logging.getLogger('name')


@leip.arg('arg', nargs=argparse.REMAINDER)
@leip.arg('name', nargs='?')
@leip.command
def run(app, args):
    if not args.name:
        if not os.path.exists('run.sh'):
            lg.error("no run.sh found")
        with open ('./run.sh') as F:
            snippet = shlex.split(F.read())
            if 'kea' in snippet[0]:
                snippet = snippet[1:] 
    else:
        snippet = app.conf.get('snippet.{}'.format(args.name))
        
    sys.argv = sys.argv[:sys.argv.index('run')] + snippet + args.arg
    kea_app = Kea()
    kea_app.run()

@leip.command
def jobrun(app, args):

    if not os.path.exists('run.sh'):
        lg.error("no run.sh found")
        exit(-1)
    with open ('./run.sh') as F:
        code = F.read()

    os.system(code)
    exit()
    
@leip.command
def jr(app, args):
    return jobrun(app, args)

    
@leip.arg('args', nargs=argparse.REMAINDER)
@leip.arg('name')
@leip.command
def snipset(app, args):
    cl = args.args
    lg.warning('saving to "%s": %s', args.name, " ".join(args.args))
    lconf = leip.get_local_config_file('kea')
    if lconf.has_key('snippet'):
        lconf['snippet'] = {}
    lconf['snippet.{}'.format(args.name)] = args.args
    leip.save_local_config_file(lconf, 'kea')

    #force rehash
    leip.get_config('kea', rehash=True)


@leip.arg('command_line', nargs=argparse.REMAINDER)
@leip.command
def jobset(app, args):
    """
    Set a local job

    if no command line is provided a prompt is provided
    """
    
    if len(args.command_line) == 0:
        import toMaKe.ui
        default = ""
        if os.path.exists('run.sh'):
            with open('run.sh') as F:
                default = F.read().strip()
        jcl = toMaKe.ui.askUser("cl", appname='kea', default=default,
                                prompt='cl: ')
    else:
        jcl = args.command_line

    lg.info('saving as job: %s', " ".join(jcl))
    with open('./run.sh', 'w') as F:
        F.write(jcl.rstrip() +  "\n")

@leip.arg('command_line', nargs=argparse.REMAINDER)
@leip.command
def js(app, args):
    """
    Alias for jobset
    """
    return jobset(app, args)
    



meta = {'database': {'category': 'db', 'pattern': '/\\S+/', 'type': 'file', 'transform': '{{file.database}}.1.ebwt'}, 'samout': {'category': 'output', 'default': '__stdout__', 'type': 'file', 'pattern': '/\\S+/'}, 'forward_reads': {'category': 'input', 'pattern': '/\\S+/', 'type': 'file'}, 'reverse_reads': {'category': 'input', 'pattern': '/\\S+/', 'type': 'file'}, 'reads': {'category': 'input', 'pattern': '/\\S+/', 'type': 'file'}, 'twoints': {'pattern': '/[0-9]+,[0-9]/', 'type': 'other'}}



import copy
import logging
import os

from jinja2 import Template
import arrow

import leip

from kea.plugin.executor_simple import BasicExecutor, get_deferred_cl
from kea.utils import get_base_uid

lg = logging.getLogger(__name__)


@leip.hook('pre_argparse')
def prep_sge_exec(app):
    if app.executor == 'pbs':
        pbs_group = app.parser.add_argument_group('PBS Executor')
        pbs_group.add_argument('--pbs_nodes',
                               help='No nodes requested', type=int)
        pbs_group.add_argument('-j', '--pbs_ppn', type=int,
                               help='No ppn requested (default=cl per job)')
        pbs_group.add_argument('--pbs_account', 
                               help='Account requested (default none)')
        pbs_group.add_argument('-w', '--walltime',
                                help=('max time that this process can take'))


PBS_SUBMIT_SCRIPT_HEADER = """#!/bin/bash
#PBS -N {{appname}}.{{uid}}
#PBS -e {{ cwd }}/{{appname}}.{{uid}}.$PBS_JOBID.err
#PBS -o {{ cwd }}/{{appname}}.{{uid}}.$PBS_JOBID.out
#PBS -l nodes={{pbs_nodes}}:ppn={{pbs_ppn}}

{% if pbs_account -%}
  #PBS -A {{ pbs_account }}{% endif %}
{% if walltime -%}
  #PBS -l walltime={{ walltime }}{% endif %}

set -v

"""


class PbsExecutor(BasicExecutor):

    def __init__(self, app):

        super(PbsExecutor, self).__init__(app)

        self.buffer = []
        self.batch = 0
        self.clno = 0


    def submit_to_pbs(self):
        uid = get_base_uid()
        
        #write pbs script
        pbs_script = '{}.{}.pbs'.format(self.app.name, uid)

        template = Template(PBS_SUBMIT_SCRIPT_HEADER)

        data = dict(self.app.defargs)
        data['appname'] = self.app.name
        data['cwd'] = os.getcwd()
        data['uid'] = uid

        
        lg.debug("submit to pbs with uid %s", uid)
        for info in self.buffer:

            info['submitted'] = arrow.utcnow()
            info['pbs_uid'] = uid
            info['pbs_script_file'] = pbs_script

            with open(pbs_script, 'w') as F:
                F.write(template.render(**data))
                for info in self.buffer:
                    F.write("( " + " ".join(get_deferred_cl(info)))
                    F.write(" ) & \n")
                F.write("wait\n")
                F.write('echo "done"\n')

            self.clno += 1

        #fire & forget the pbs job
        pbs_cl = ['qsub', pbs_script]
        print " ".join(pbs_cl)
        self.buffer = []
        self.batch += 1


    def fire(self, info):
        self.app.run_hook('pre_fire', info)
        self.buffer.append(copy.copy(info))
        info['returncode'] = 0
        info['status'] = 'queued'
        self.app.run_hook('post_fire', info)
        if len(self.buffer) >= self.app.defargs['pbs_ppn']:
            lg.info("submitting pbs job. No commands: %d", len(self.buffer))
            self.submit_to_pbs()

    def finish(self):
        if len(self.buffer) > 0:
            lg.info("submitting pbs job. No commands: %d", len(self.buffer))
            self.submit_to_pbs()


conf = leip.get_config('kea')
conf['executors.pbs'] = PbsExecutor

